//==============================================================================
//!	@file   my_go.cpp
//!	@brief  囲碁AI
//==============================================================================

//
//	Cedec Go twenty ELEven 
//

/*

フレームワーク
	囲碁プログラム 勝也 と コンピュータ囲碁のページ 
	http://homepage1.nifty.com/Ike/katsunari/

	囲碁プログラムの作り方(基本編)
	http://homepage1.nifty.com/Ike/katsunari/howto_kihon.html

	通信対戦対応サンプルについて 
	http://www.computer-go.jp/gifu2004/regulations/sample.html

その他資料
	YSSと彩のページ 
	http://www32.ocn.ne.jp/~yss/index_j.html

	GTP - Go Text Protocol 
	http://www.lysator.liu.se/~gunnar/gtp/

*/

//==============================================================================
//  Include
//==============================================================================
#include "stdafx.h"
#include "my_go.h"
#include <stdarg.h>
#include <cassert>
#include <ctime>
#include <algorithm>

//==============================================================================
//  etc
//==============================================================================

	//--------------------------------------S
	//! @brief デバッグ用ロガー
	//--------------------------------------
#ifdef _DEBUG
//	#define DEBUGPRINT(...)		((void)0)
	#define DEBUGPRINT			g_DebugLogger->printf
	class DebugLogger{
		FILE *m_fp;
		static const int BUFFER_SIZE = 4096;
		static char m_buffer[BUFFER_SIZE];
	public:
		DebugLogger( std::string filename )
		{
			//fopen_s( &m_fp,  filename.c_str(), "wt" );
			m_fp = _fsopen( filename.c_str(), "wt",_SH_DENYNO );
			//m_fp = fopen( filename.c_str(), "wt" );
		}
		~DebugLogger()
		{
			if( m_fp ){
				fclose( m_fp );
			}
		}
		void printf(char*fmt,...)
		{
			if( m_fp ){
				va_list	ap;
				va_start(ap, fmt);
				vsnprintf_s( m_buffer, BUFFER_SIZE, BUFFER_SIZE-1, fmt, ap);
				va_end(ap);

				fputs( m_buffer, m_fp );
				fflush(m_fp);
			}
		}
	};
	char DebugLogger::m_buffer[DebugLogger::BUFFER_SIZE];
	static DebugLogger * g_DebugLogger = NULL;
#else
	#define DEBUGPRINT(...)		((void)0)
#endif

	//--------------------------------------S
	//! @brief  COLOR列挙値に対応する文字列を返す
	//--------------------------------------
	const char * colorstr( COLOR color )
	{
		static const char * const tbl[] = {
			"EMPTY",
			"WHITE",
			"BLACK",
			"BLACK_TERRITORY",
			"WHITE_TERRITORY",
			"ALIVE",
			"ERROR",
		};
		return tbl[color];
	}
	const char * colorstr_short( COLOR color )
	{
		static const char * const tbl[] = {
			"□",
			"白",
			"黒",
			"b ",
			"w ",
			"A ",
			"E ",
		};
		return tbl[color];
	}

//==============================================================================
//  engine::impl
//==============================================================================
namespace my_go {

//! 碁盤の大きさ（盤外を含む） 
static const int BOARD_MAXSIZE = (19+2);	

//! 碁盤の表示のための文字
static const char stone[4] = { '+', 'b', 'w', '?' };

//! 着手可能点の要素構造体
struct point{
	int x;
	int y;
	point( int a_x, int a_y )
	: x(a_x)
	, y(a_y)
	{
	}
};

//! 着手可能点の要素構造体
struct point_score{
	int x;
	int y;
	int s;
	point_score( int a_x, int a_y )
	: x(a_x)
	, y(a_y)
	, s(0)
	{
	}
};
bool point_score_func_big( const point_score &a, const point_score &b )
{
	return a.s > b.s;
}

class engine::Impl{

public:
	enum STATE{
		STATE_SPACE=0,	           //!< 空点(石が置かれていない)
		STATE_BLACK=1,	           //!< 黒   
		STATE_WHITE=2,	           //!< 白   
		STATE_OUT  =3,	           //!< 盤外 
	};

private:
	//! 棋譜 
	struct type_log{
		STATE color;
		int x;
		int y;

		//以下undo用の拡張
		//メモリバカ食いで、計算で求めるもっといい方法があるとおもうけど時間がないので応急的に

		//! アゲハマ 
		int black_prisoner;
		int white_prisoner;

		//! 劫の位置 
		int ko_x;
		int ko_y;

		//! 劫が発生した手数
		int ko_num;

		//! 碁盤
		STATE board[BOARD_MAXSIZE][BOARD_MAXSIZE];	

		type_log( 
			STATE a_color, 
			int a_x, 
			int a_y, 
			int a_black_prisoner,
			int a_white_prisoner,
			int a_ko_x,
			int a_ko_y,
			int a_ko_num,
			STATE a_board[BOARD_MAXSIZE][BOARD_MAXSIZE] )
		: color( a_color )
		, x(a_x)
		, y(a_y)
		, black_prisoner(a_black_prisoner)
		, white_prisoner(a_white_prisoner)
		, ko_x(a_ko_x)
		, ko_y(a_ko_y)
		, ko_num(a_ko_num)
		{
			for(int yy=0;yy<BOARD_MAXSIZE;yy++){
				for(int xx=0;xx<BOARD_MAXSIZE;xx++){
					board[yy][xx] = a_board[yy][xx];
				}
			}
		}
	};

	//! 碁盤
	STATE m_board[BOARD_MAXSIZE][BOARD_MAXSIZE];
	int m_boardsize;

	//! アゲハマ 
	int m_black_prisoner;
	int m_white_prisoner;

	//! 劫の位置 
	int m_ko_x;
	int m_ko_y;

	//! 劫が発生した手数
	int m_ko_num;

	//! 棋譜
	std::vector<type_log> m_logData;

	//! 合法手かどうか調べるのに使う
	bool m_checkBoard[BOARD_MAXSIZE][BOARD_MAXSIZE];

public:

private:
	//--------------------------------------
	//! 相手の色 を計算する
	//--------------------------------------
	static inline STATE reverseColor( STATE color )
	{
		return ( color == STATE_BLACK ? STATE_WHITE : STATE_BLACK );
	}

	//--------------------------------------
	//! @brief  手数を取得する
	//--------------------------------------
	unsigned GetMove() const 
	{
		return (m_logData.size() + 1);
	}

public:

	//--------------------------------------
	//! @brief  碁盤の初期化                
	//--------------------------------------
	void InitializeBoard( int size, float komi )
	{
		int x, y;
		
		m_boardsize = size + 2; //盤外も含む
		assert(m_boardsize <= BOARD_MAXSIZE);

		//まず全部盤外でクリアする
		for( y=0; y < (m_boardsize); y++ ){
			for( x=0; x < (m_boardsize); x++ ){
				m_board[y][x] = STATE_OUT;
			}
		}

		//次に、置けるところを空白でクリア
		for( y=1; y < (m_boardsize-1); y++ ){
			for( x=1; x < (m_boardsize-1); x++ ){
				m_board[y][x] = STATE_SPACE;
			}
		}

		// アゲハマ 
		m_black_prisoner = 0;
		m_white_prisoner = 0;

		//ログのクリア
		m_logData.clear();
		m_logData.reserve(500);

		//乱数初期化
		srand( static_cast<unsigned int>( time(NULL) ) );
	}

	//--------------------------------------
	//! @brief  碁盤を表示する
	//--------------------------------------
	/*
	void DisplayBoard( void )
	{
		int x, y;

		printf( "    [ 1 2 3 4 5 6 7 8 910111213141516171819]\n" );
		for( y=1; y < (BOARD_MAXSIZE-1); y++ ){
			printf( "[%2d] ", y );
			for( x=1; x < (BOARD_MAXSIZE-1); x++ ){
				printf( " %c", stone[m_board[y][x]] );
			}
			printf( "\n" );
		}
	}
	*/

	//--------------------------------------
	//! @brief  着手位置の入力
	//--------------------------------------
	bool ThinkMove( STATE color, int *x, int *y )
	{
		//直前の相手がパスしているならパス
		//! @note コンピュータ以後では2回連続でパスしたら終了
		//! @todo これだと相手がパスさえすればパスし返すので、正しい条件がいるとおもう
		if( m_logData.size() > 0 ){
			const type_log & rlog = m_logData.at(m_logData.size()-1);
			if( rlog.x < 1 || rlog.y < 1 ){
				return false;
			}
		}
		
		//左上から順に置けるとこに置く
#if 0
		{
			for( int xx=1; xx < (m_boardsize-1); xx++ ){
				for( int yy=1; yy< (m_boardsize-1); yy++ ){
					if( CheckLegal( color, xx, yy ) ){
						*x = xx;
						*y = yy;
						return true;
					}

				}
			}
		}
#endif
		//ランダムに置けるとこに置く
#if 0
		{
			//置けるところのリストを作る
			std::vector<point> m_legal_points;
			m_legal_points.clear();
			m_legal_points.reserve(m_boardsize*m_boardsize);
			for( int xx=1; xx < (m_boardsize-1); xx++ ){
				for( int yy=1; yy< (m_boardsize-1); yy++ ){
					if( CheckLegal( color, xx, yy ) ){
						m_legal_points.push_back( point(xx,yy) );
					}
				}
			}
			
			//リストからランダムに選ぶ
			//! @todo ここをちゃんと各点の有効性を評価するようにすればＡＩらしく・・・
			if( m_legal_points.size() > 0 ){
				int index = rand() % m_legal_points.size();
				point &r = m_legal_points.at(index);
				*x = r.x;
				*y = r.y;
				return true;
			}
		}
#endif	
		//着手可能な各点に点数をつけ、最高点の着手可能点を選ぶ
#if 1
		{
			//最初の手番のときは４つの星目のいずれかが空いてるならそこにうつ
			if( m_logData.size() == 0 && color == STATE_BLACK ||
				m_logData.size() == 1 && color == STATE_WHITE  )
			{
				if( CheckLegal( color, 3, 3 ) ){
					DEBUGPRINT("first attack!\n");
					*x = 3;
					*y = 3;
					return true;
				}
				if( CheckLegal( color, 7, 7 ) ){
					DEBUGPRINT("first attack!\n");
					*x = 7;
					*y = 7;
					return true;
				}
				if( CheckLegal( color, 3, 7 ) ){
					DEBUGPRINT("first attack!\n");
					*x = 3;
					*y = 7;
					return true;
				}
				if( CheckLegal( color, 7, 3 ) ){
					DEBUGPRINT("first attack!\n");
					*x = 7;
					*y = 3;
					return true;
				}
			}

			//着手可能点リストの作成
			std::vector<point_score> m_legal_points;
			m_legal_points.clear();
			m_legal_points.reserve(m_boardsize*m_boardsize);
			for( int xx=1; xx < (m_boardsize-1); xx++ ){
				for( int yy=1; yy< (m_boardsize-1); yy++ ){
					if( CheckLegal( color, xx, yy ) ){
						m_legal_points.push_back( point_score(xx,yy) );
					}
				}
			}
			DEBUGPRINT("legal_points = %d\n", m_legal_points.size() );

			//着手可能点がなければパスをする
			if( m_legal_points.size() == 0 )
			{
				DEBUGPRINT("pass\n");
				return false;
			}

			//着手可能点１つなら迷う余地がない
			if( m_legal_points.size() == 1 )
			{
				DEBUGPRINT("point only one\n");
				const point_score &r = m_legal_points.at(0);
				*x = r.x;
				*y = r.y;
				return true;
			}

			//着手可能点リストの各点にスコアをつける
			Calculate_Score( color, m_legal_points);

			//デバッグ表示
			DEBUGPRINT("\tlegal_points\n");
			for( int i=0; i<m_legal_points.size(); i++ ){
				DEBUGPRINT("\t x=%d,y=%d,s=%d\n",m_legal_points[i].x,m_legal_points[i].y,m_legal_points[i].s);
			}

			//着手可能点リストをスコアの高い順にソートする
			std::sort( m_legal_points.begin(), m_legal_points.end(), point_score_func_big );

			//デバッグ表示
			DEBUGPRINT("\tlegal_points\n");
			for( int i=0; i<m_legal_points.size(); i++ ){
				DEBUGPRINT("\t x=%d,y=%d,s=%d\n",m_legal_points[i].x,m_legal_points[i].y,m_legal_points[i].s);
			}

			//最高スコアの着手可能点を選択する
			{
				//最高スコアの着手可能点の数を数える
				int max_score = m_legal_points.at(0).s;
				int max_score_count = 1;
				for( unsigned i=1; i<m_legal_points.size(); i++ ){
					//最高スコアがなくなったら抜ける
					if( max_score > m_legal_points.at(i).s ){
						break;
					}
					//カウントアップ
					++max_score_count;
				}

				//着手可能点が複数あるならランダムに選ぶ
				int index = (rand()/10) % max_score_count;
				const point_score &r = m_legal_points.at(index);
				*x = r.x;
				*y = r.y;
				return true;
			}
		}
#endif
		return false;
	}

	//--------------------------------------
	//! @brief  盤状態取得
	//--------------------------------------
	STATE GetState( int x, int y ) const
	{
		return m_board[y][x];
	}
	
	//--------------------------------------
	//! @brief  合法手かどうか調べる
	//--------------------------------------
	bool CheckLegal( STATE color, int x, int y )
	{
		/* 空点じゃないと置けません */
		if( m_board[y][x] != STATE_SPACE ){
			return( false );
		}

		/* 一手前に劫を取られていたら置けません */
		if( GetMove() > 1 ){
			if( m_ko_x == x && m_ko_y == y && m_ko_num == (GetMove()-1) ){
				return( false );
			}
		}

		/* 自殺手なら置けません */
		if( CheckSuicide( color, x, y ) ){
			return( false );
		}

		/* 以上のチェックをすべてクリアできたので置けます */
		return( true );
	}

	//--------------------------------------
	//! @brief 碁盤に石を置く                       
	//--------------------------------------
	void SetStone( STATE color, int x, int y )
	{
		int prisonerN;    /* 取り除かれた石の数（上） */
		int prisonerE;    /* 取り除かれた石の数（右） */
		int prisonerS;    /* 取り除かれた石の数（下） */
		int prisonerW;    /* 取り除かれた石の数（左） */
		int prisonerAll;  /* 取り除かれた石の総数 */
		bool koFlag;       /* 劫かどうか */

		/* 座標(x,y)に石を置く */
		m_board[y][x] = color;

		/* 置いた石の隣に同じ色の石はあるか？ */
		if( m_board[y+1][x] != color &&
			m_board[y-1][x] != color &&
			m_board[y][x+1] != color &&
			m_board[y][x-1] != color ){
			/* 同じ色の石がないならば劫かもしれない */
			koFlag = true;
		}else{
			/* 同じ色の石があるならば劫ではない */
			koFlag = false;
		}

		/* 取り除かれた石の数 */
		prisonerN = prisonerE = prisonerS = prisonerW = 0;

		/* 置いた石の周囲の相手の石が死んでいれば碁盤から取り除く */
		if( y > 1 ){
			prisonerN = RemoveStone( m_board, color, x, y-1 );
		}
		if( x > 1 ){
			prisonerW = RemoveStone( m_board, color, x-1, y );
		}
		if( y < (m_boardsize-2) ){
			prisonerS = RemoveStone( m_board, color, x, y+1 );
		}
		if( x < (m_boardsize-2) ){
			prisonerE = RemoveStone( m_board, color, x+1, y );
		}

		/* 取り除かれた石の総数 */
		prisonerAll = prisonerN + prisonerE + prisonerS + prisonerW;

		/* 置いた石の隣に同じ色の石がなく、取り除かれた石も１つならば劫 */
		if( koFlag && prisonerAll == 1 ){
			/* 劫の発生した手数を覚える */
			m_ko_num = GetMove();
			/* 劫の座標を覚える */
			if( prisonerE == 1 ){
				/* 取り除かれた石が右 */
				m_ko_x = x+1;
				m_ko_y = y;
			}else if( prisonerS == 1 ){
				/* 取り除かれた石が下 */
				m_ko_x = x;
				m_ko_y = y+1;
			}else if( prisonerW == 1 ){
				/* 取り除かれた石が左 */
				m_ko_x = x-1;
				m_ko_y = y;
			}else if( prisonerN == 1 ){
				/* 取り除かれた石が上 */
				m_ko_x = x;
				m_ko_y = y-1;
			}
		}

		/* アゲハマの更新 */
		if( prisonerAll > 0 ){
			if( color == STATE_BLACK ){
				m_black_prisoner += prisonerAll;
			}else if( color == STATE_WHITE ){
				m_white_prisoner += prisonerAll;
			}
		}
	}

	//--------------------------------------
	//! 棋譜に記録                                                       
	//--------------------------------------
	void RecordMove( STATE color, int x, int y )
	{
		m_logData.push_back( 
			type_log(color,x,y,m_black_prisoner,
			m_white_prisoner,m_ko_x,m_ko_y,m_ko_num,m_board) );
	}

	//--------------------------------------
	//! 棋譜を戻す
	//--------------------------------------
	void UndoMove()
	{
		if( m_logData.size() > 0 ){
			const type_log & rlog = m_logData.at(m_logData.size()-1);
			
			//書き戻す
			m_black_prisoner = rlog.black_prisoner;
			m_white_prisoner = rlog.white_prisoner;
			m_ko_x = rlog.ko_x;
			m_ko_y = rlog.ko_y;
			m_ko_num = rlog.ko_num;
			
			for(int yy=0;yy<BOARD_MAXSIZE;yy++){
				for(int xx=0;xx<BOARD_MAXSIZE;xx++){
					m_board[yy][xx] = rlog.board[yy][xx];
				}
			}
			
			//１つ減らす
			m_logData.pop_back();
		}
	}

private:

	//--------------------------------------
	//! @brief 引数で受け取ったテスト用碁盤に石を置く                       
	//--------------------------------------
	void SetStoneForTest( STATE board[][BOARD_MAXSIZE], STATE color, int x, int y ) 
	{
		board[y][x] = color;

		if( y > 1 ){
			RemoveStone( board, color, x, y-1 );
		}
		if( x > 1 ){
			RemoveStone( board, color, x-1, y );
		}
		if( y < (m_boardsize-2) ){
			RemoveStone( board, color, x, y+1 );
		}
		if( x < (m_boardsize-2) ){
			RemoveStone( board, color, x+1, y );
		}
	}

	//--------------------------------------
	//! @brief  盤面評価用ラッパー
	//--------------------------------------
	void AddValue( int valueboard[][BOARD_MAXSIZE], int x, int y, int value ) const
	{
		if( x > 0 && x < (m_boardsize-1) && y > 0 && y < (m_boardsize-1) ){
			valueboard[y][x] += value;
		}
	}

	//--------------------------------------
	//! @brief  現在の盤面の評価をする
	//--------------------------------------
	int EvaluateBoard( const STATE board[][BOARD_MAXSIZE], STATE color ) const
	{
		static int	 blackBoard[BOARD_MAXSIZE][BOARD_MAXSIZE];
		static int	 whiteBoard[BOARD_MAXSIZE][BOARD_MAXSIZE];
		static STATE totalBoard[BOARD_MAXSIZE][BOARD_MAXSIZE];

		DEBUGPRINT("\tboard\n\t");
		  	for( int x=0; x < (m_boardsize-2); x++ ){
	  	for( int y=0; y < (m_boardsize-2); y++ ){
				DEBUGPRINT("%c,", stone[board[y+1][x+1]] );
			}
			DEBUGPRINT("\n\t");
		}
		DEBUGPRINT("\n");
		
		//初期化する
		for( int y=1; y < (m_boardsize-1); y++ ){
    		for( int x=1; x < (m_boardsize-1); x++ ){
				blackBoard[y][x] = 0;
                whiteBoard[y][x] = 0;
    			totalBoard[y][x] = STATE_SPACE;
    		}
		}

		//現在の勢力図の作成
		for( int y=1; y < (m_boardsize-1); y++ ){
    		for( int x=1; x < (m_boardsize-1); x++ ){
				if( board[y][x] == STATE_BLACK ){
					AddValue( blackBoard, x-3, y  , 9 );
					AddValue( blackBoard, x-2, y-1, 9 );
					AddValue( blackBoard, x-2, y  , 11 );
					AddValue( blackBoard, x-2, y+1, 9 );
					AddValue( blackBoard, x-1, y-2, 10 );
					AddValue( blackBoard, x-1, y-1, 11 );
					AddValue( blackBoard, x-1, y  , 10 );
					AddValue( blackBoard, x-1, y+1, 11 );
					AddValue( blackBoard, x-1, y+2, 9 );
					AddValue( blackBoard, x,   y-3, 9 );
					AddValue( blackBoard, x,   y-2, 11 );
					AddValue( blackBoard, x,   y-1, 10 );
					AddValue( blackBoard, x,   y  , 10 );
					AddValue( blackBoard, x,   y+1, 10 );
					AddValue( blackBoard, x,   y+2, 11 );
					AddValue( blackBoard, x,   y+3, 9 );
					AddValue( blackBoard, x+1, y-2, 9 );
					AddValue( blackBoard, x+1, y-1, 11 );
					AddValue( blackBoard, x+1, y  , 10 );
					AddValue( blackBoard, x+1, y+1, 11 );
					AddValue( blackBoard, x+1, y+2, 9 );
					AddValue( blackBoard, x+2, y-1, 9 );
					AddValue( blackBoard, x+2, y  , 11 );
					AddValue( blackBoard, x+2, y+1, 9 );
					AddValue( blackBoard, x+3, y  , 9 );
				}
				if( board[y][x] == STATE_WHITE ){
					AddValue( whiteBoard, x-3, y  , 9 );
					AddValue( whiteBoard, x-2, y-1, 9 );
					AddValue( whiteBoard, x-2, y  , 11 );
					AddValue( whiteBoard, x-2, y+1, 9 );
					AddValue( whiteBoard, x-1, y-2, 9 );
					AddValue( whiteBoard, x-1, y-1, 11 );
					AddValue( whiteBoard, x-1, y  , 10 );
					AddValue( whiteBoard, x-1, y+1, 11 );
					AddValue( whiteBoard, x-1, y+2, 9 );
					AddValue( whiteBoard, x,   y-3, 9 );
					AddValue( whiteBoard, x,   y-2, 11 );
					AddValue( whiteBoard, x,   y-1, 10 );
					AddValue( whiteBoard, x,   y  , 10 );
					AddValue( whiteBoard, x,   y+1, 10 );
					AddValue( whiteBoard, x,   y+2, 11 );
					AddValue( whiteBoard, x,   y+3, 9 );
					AddValue( whiteBoard, x+1, y-2, 9 );
					AddValue( whiteBoard, x+1, y-1, 11 );
					AddValue( whiteBoard, x+1, y  , 10 );
					AddValue( whiteBoard, x+1, y+1, 11 );
					AddValue( whiteBoard, x+1, y+2, 9 );
					AddValue( whiteBoard, x+2, y-1, 9 );
					AddValue( whiteBoard, x+2, y  , 11 );
					AddValue( whiteBoard, x+2, y+1, 9 );
					AddValue( whiteBoard, x+3, y  , 9 );
				}
			}
		}

		DEBUGPRINT("\tblack board\n\t");
		  	for( int x=0; x < (m_boardsize-2); x++ ){
	  	for( int y=0; y < (m_boardsize-2); y++ ){
				DEBUGPRINT("%3d,", blackBoard[y+1][x+1] );
			}
			DEBUGPRINT("\n\t");
		}
		DEBUGPRINT("\n");

		DEBUGPRINT("\twhite board\n\t");
		  	for( int x=0; x < (m_boardsize-2); x++ ){
	  	for( int y=0; y < (m_boardsize-2); y++ ){
				DEBUGPRINT("%3d,", whiteBoard[y+1][x+1] );
			}
			DEBUGPRINT("\n\t");
		}
		DEBUGPRINT("\n");

		int value = 0;
	  	for( int y=1; y < (m_boardsize-1); y++ ){
	  	  	for( int x=1; x < (m_boardsize-1); x++ ){
	  	    	if( board[y][x] == STATE_BLACK ){
	  	    	  	totalBoard[y][x] = STATE_BLACK;
	  	    	}else if( board[y][x] == STATE_WHITE ){
	  	     		totalBoard[y][x] = STATE_WHITE;
	  	    	}else{
					if( color == STATE_BLACK ){
	  	      			if( blackBoard[y][x] > whiteBoard[y][x] ){
		  	        		totalBoard[y][x] = STATE_BLACK;
	  		        		value++;
	  	    	  		}else if( blackBoard[y][x] < whiteBoard[y][x] ){
		  	        		totalBoard[y][x] = STATE_WHITE;
	  		        		value--;
	  	    	  		}
					}else{
	  	      			if( blackBoard[y][x] > whiteBoard[y][x] ){
		  	        		totalBoard[y][x] = STATE_BLACK;
	  		        		value--;
	  	    	  		}else if( blackBoard[y][x] < whiteBoard[y][x] ){
		  	        		totalBoard[y][x] = STATE_WHITE;
	  		        		value++;
	  	    	  		}
					}
	  	    	}
	  	  	}
	  	}
		
		DEBUGPRINT("\ttotal board\n\t");
		  	for( int x=0; x < (m_boardsize-2); x++ ){
	  	for( int y=0; y < (m_boardsize-2); y++ ){
				DEBUGPRINT("%c,", stone[totalBoard[y+1][x+1]] );
			}
			DEBUGPRINT("\n\t");
		}
		DEBUGPRINT("\n");

	 	return( value );
	}
	
	//--------------------------------------
	//! @brief  着手可能点のスコア計算
	//--------------------------------------
	void Calculate_Score( STATE color, std::vector<point_score>& legal_points ) 
	{
		DEBUGPRINT("Calculate_Score >>> \n");

		//デバッグ表示
	//	for( int i=0; i<m_legal_points.size(); i++ ){
	//		DEBUGPRINT("x=%d,y=%d,s=%d\n",m_legal_points[i].x,m_legal_points[i].y,m_legal_points[i].s);
	//	}

		//現在の盤面の評価点を取得
		DEBUGPRINT("\t--------\n" );
		DEBUGPRINT("\ttest Now\n" );
		int ScoreNow = EvaluateBoard(m_board, color);
		DEBUGPRINT("\tvalue=%d\n", ScoreNow );

		//着手可能点それぞれについて、打った場合の評価点を集める
		std::vector<point_score>::iterator p = legal_points.begin();
		while( p != legal_points.end() )
		{
			//評価用のテンポラリ盤面に現在の盤面をコピーする
			static STATE tempboard[BOARD_MAXSIZE][BOARD_MAXSIZE];
			std::memcpy( tempboard, m_board, sizeof(tempboard) );
			//テンポラリ盤面に１手打つ
			SetStoneForTest( tempboard,color, p->x, p->y );
			//打った結果の盤面を評価する
			DEBUGPRINT("\t--------\n" );
			DEBUGPRINT("\ttest x=%d,y=%d\n", p->x,p->y );
			p->s = EvaluateBoard( tempboard, color );
			DEBUGPRINT("\tvalue=%d\n", p->s);

	//おためし		
	/*
			//すでに大きく囲まれているところなら減点する
			ClearCheckBoard();
			if( DoCheckRemoveStone( tempboard, color, p->x, p->y ) )
			{
				DEBUGPRINT("\tcheck remove\n");
				p->s -= 30;
			}
	*/
	//おためしend

			++p;
		}
	}

	//--------------------------------------
	//! @brief  自殺手かどうか調べる                                             
	//--------------------------------------
	bool CheckSuicide( STATE color, int x, int y )
	{
		bool rtnVal;
		STATE opponent;  /* 相手の色 */

		/* 仮に石を置く */
		m_board[y][x] = color;

		/* マークのクリア */
		ClearCheckBoard();

		/* その石は相手に囲まれているか調べる */
		rtnVal = DoCheckRemoveStone( m_board, color, x, y );

		/* 囲まれているならば自殺手の可能性あり */
		if( rtnVal ){

			/* 相手の色を求める */
			opponent = reverseColor( color );

			/* その石を置いたことにより、隣の相手の石が取れるなら自殺手ではない */
			if( x > 1 ){
				/* 隣は相手？ */
				if( m_board[y][x-1] == opponent ){
					/* マークのクリア */
					ClearCheckBoard();
					/* 相手の石は囲まれているか？ */
					rtnVal = DoCheckRemoveStone( m_board, opponent, x-1, y );
					/* 相手の石を取れるので自殺手ではない */
					if( rtnVal ){
						/* 盤を元に戻す */
						m_board[y][x] = STATE_SPACE;
						return( false );
					}
				}
			}
			if( y > 1 ){
				/* 隣は相手？ */
				if( m_board[y-1][x] == opponent ){
					/* マークのクリア */
					ClearCheckBoard();
					/* 相手の石は囲まれているか？ */
					rtnVal = DoCheckRemoveStone( m_board, opponent, x, y-1 );
					/* 相手の石を取れるので自殺手ではない */
					if( rtnVal ){
						/* 盤を元に戻す */
						m_board[y][x] = STATE_SPACE;
						return( false );
					}
				}
			}
			if( x < (m_boardsize-2) ){
				/* 隣は相手？ */
				if( m_board[y][x+1] == opponent ){
					/* マークのクリア */
					ClearCheckBoard();
					/* 相手の石は囲まれているか？ */
					rtnVal = DoCheckRemoveStone( m_board,opponent, x+1, y );
					/* 相手の石を取れるので自殺手ではない */
					if( rtnVal ){
						/* 盤を元に戻す */
						m_board[y][x] = STATE_SPACE;
						return( false );
					}
				}
			}
			if( y < (m_boardsize-2) ){
				/* 隣は相手？ */
				if( m_board[y+1][x] == opponent ){
					/* マークのクリア */
					ClearCheckBoard();
					/* 相手の石は囲まれているか？ */
					rtnVal = DoCheckRemoveStone( m_board,opponent, x, y+1 );
					/* 相手の石を取れるので自殺手ではない */
					if( rtnVal ){
						/* 盤を元に戻す */
						m_board[y][x] = STATE_SPACE;
						return( false );
					}
				}
			}

			/* 盤を元に戻す */
			m_board[y][x] = STATE_SPACE;

			/* 相手の石を取れないなら自殺手 */
			return( true );

		}else{

			/* 盤を元に戻す */
			m_board[y][x] = STATE_SPACE;

			/* 囲まれていないので自殺手ではない */
			return( false );
		}
	}

	//--------------------------------------
	//! @brief チェック用の碁盤をクリア
	//--------------------------------------
	void ClearCheckBoard( void )
	{
		int x, y;

		for( y=1; y < (m_boardsize-1); y++ ){
			for( x=1; x < (m_boardsize-1); x++ ){
				m_checkBoard[y][x] = false;
			}
		}
	}

	//--------------------------------------
	//! @brief 座標(x,y)にあるcolor石が相手に囲まれているか調べる
	//! @ret   空点があればfalseを返し、空点がなければtrueを返す
	//--------------------------------------
	bool DoCheckRemoveStone( STATE board[][BOARD_MAXSIZE], STATE color, int x, int y )
	{
		int rtn;

		/* その場所は既に調べた点ならおしまい */  
		if( m_checkBoard[y][x] ){
			return( true );
		}

		/* 調べたことをマークする */
		m_checkBoard[y][x] = true;

		/* 何も置かれていないならばおしまい */
		if( board[y][x] == STATE_SPACE ){
			return( false );
		}

		/* 同じ色の石ならばその石の隣も調べる */  
		if( board[y][x] == color ){
			/* その石の左(x-1,y)を調べる */
			if( x > 1 ){
				rtn = DoCheckRemoveStone( board, color, x-1, y );
				if( !rtn ){
					return( false );
				}
			}
			/* その石の上(x,y-1)を調べる */
			if( y > 1 ){
				rtn = DoCheckRemoveStone( board, color, x, y-1 );
				if( !rtn ){
					return( false );
				}
			}
			/* その石の右(x+1,y)を調べる */
			if( x < (m_boardsize-2) ){
				rtn = DoCheckRemoveStone( board, color, x+1, y );
				if( !rtn ){
					return( false );
				}
			}
			/* その石の下(x,y+1)を調べる */
			if( y < (m_boardsize-2) ){
				rtn = DoCheckRemoveStone( board, color, x, y+1 );
				if( !rtn ){
					return( false );
				}
			}
		}

		/* 相手の色の石があった */  
		return( true );
	}

	//--------------------------------------
	//! @brief 勝敗の判定
	//	このプログラムでは地を数えていません     
	//	アゲハマの多い方を勝ちと判定しています   
	//--------------------------------------
#if 0
	int CountScore()
	{
		int black_score;
		int white_score;

		/* 碁盤上の地を数えるべきところだが省略 */
		black_score = 0;
		white_score = 0;

		/* アゲハマを加える */
		black_score += m_black_prisoner;
		white_score += m_white_prisoner;

		/* 黒－白を返す */
		return( black_score - white_score );
	}
#endif

	//--------------------------------------
	//! @brief 座標(x,y)の石が死んでいれば碁盤から取り除く
	//--------------------------------------
	int	RemoveStone( STATE board[][BOARD_MAXSIZE], STATE color, int x, int y )
	{
		int prisoner;  /* 取り除かれた石数 */

		/* 置いた石と同じ色なら取らない */
		if( board[y][x] == color ){
			return( 0 );
		}

		/* 空点なら取らない */
		if( board[y][x] == STATE_SPACE ){
			return( 0 );
		}

		/* マークのクリア */
		ClearCheckBoard();

		/* 囲まれているなら取る */
		if( DoCheckRemoveStone( board, board[y][x], x, y ) ){
			prisoner = DoRemoveStone( board, board[y][x], x, y, 0 );
			return( prisoner );
		}

		return( 0 );
	}

	//--------------------------------------
	//! @brief 座標(x,y)のcolor石を碁盤から取り除き、取った石の数を返す
	//--------------------------------------
	int DoRemoveStone( STATE board[][BOARD_MAXSIZE], STATE color, int x, int y, int prisoner ) const
	{
		/* 取り除かれる石と同じ色ならば石を取る */
		if( board[y][x] == color ){

			/* 取った石の数を１つ増やす */
			prisoner++;

			/* その座標に空点を置く */
			board[y][x] = STATE_SPACE;

			/* 左を調べる */
			if( x > 1 ){
				prisoner = DoRemoveStone( board, color, x-1, y, prisoner );
			}
			/* 上を調べる */
			if( y > 1 ){
				prisoner = DoRemoveStone( board, color, x, y-1, prisoner );
			}
			/* 右を調べる */
			if( x < (m_boardsize-2) ){
				prisoner = DoRemoveStone( board, color, x+1, y, prisoner );
			}
			/* 下を調べる */
			if( y < (m_boardsize-2) ){
				prisoner = DoRemoveStone( board, color, x, y+1, prisoner );
			}
		}

		/* 取った石の数を返す */
		return( prisoner );
	}

};
};

//==============================================================================
//  engine
//==============================================================================
namespace my_go {
	
	inline 
	engine::Impl::STATE COLOR2STATE(COLOR color)
	{
		switch( color ){
		case BLACK:
			return engine::Impl::STATE_BLACK;
		case WHITE:
			return engine::Impl::STATE_WHITE;
		default:
			assert( !"COLOR2STATE illegal color" );
			return engine::Impl::STATE_BLACK;
		}
	}

	//--------------------------------------
	//! @brief  コンストラクタ
	//--------------------------------------
	engine::engine()
	{
		m_pImpl = new Impl();
		assert( m_pImpl );
		
#ifdef _DEBUG
		assert(!g_DebugLogger );
		g_DebugLogger = new DebugLogger("testlog.txt");
		assert( g_DebugLogger );
#endif	
		DEBUGPRINT("--- engine::engine()\n");
	}


	//--------------------------------------
	//! @brief  デストラクタ
	//--------------------------------------
	engine::~engine()
	{
		DEBUGPRINT("--- engine::~engine()\n");
		
		assert( m_pImpl );
		delete m_pImpl;
		
#ifdef _DEBUG
		assert( g_DebugLogger );
		delete g_DebugLogger;
#endif
	}
#if 0
	//--------------------------------------
	//! @brief  盤面情報出力
	//--------------------------------------
	/*
	void engine::dumpboard() 
	{
		if( !m_DebugFileStream ){
			return;
		}

		for(int ii=0; ii < m_Size; ii++) {
			for (int jj=0; jj < m_Size; jj++) {
				fprintf( m_DebugFileStream, colorstr_short(m_Board[ii][jj])) ;
			}
			fprintf( m_DebugFileStream, "\n" );
		}
		fprintf( m_DebugFileStream,"\n" );
	}
	*/
#endif
	//--------------------------------------
	//! @brief  リセット
	//--------------------------------------
	void engine::reset(int size, float komi) 
	{
		DEBUGPRINT("--- engine::reset(size=%d,komi=%f)\n", size, komi );
		assert( m_pImpl );
		
		/* 碁盤の初期化 */
		m_pImpl->InitializeBoard(size,komi);
	}

	//--------------------------------------
	//! @brief  思考処理関数
	//--------------------------------------
	void engine::genmove(int& i, int& j, COLOR color)
	{
		DEBUGPRINT("--- engine::genmove(color=%s)\n", colorstr(color) );
		assert( m_pImpl );

		// 着手位置を入力 
		int xThink, yThink; 
		bool result = m_pImpl->ThinkMove( COLOR2STATE(color), &xThink, &yThink);

		// 投了またはパスかどうか
		if( result ){
			i = xThink - 1;
			j = yThink - 1;
			DEBUGPRINT("continue i=%d, j=%d\n", i, j );
		}else{
			i = -1;
			j = -1;
			DEBUGPRINT("pass\n" );
		}
	}
		
	//--------------------------------------
	//! @brief  そこに石を置けるかどうか
	//--------------------------------------
	bool engine::legal_move(int i, int j, COLOR color)
	{
		DEBUGPRINT("--- engine::legal_move(i=%d,j=%d,color=%s)\n", i, j, colorstr(color) );
		assert( m_pImpl );
		
		//パスなら
		if( i<0 || j<0 ){
			DEBUGPRINT("pass\n");
			return true;
		}

		//石を置けるか調べる
		bool ret = m_pImpl->CheckLegal( COLOR2STATE(color), i+1, j+1 );
		if( ret ){
			DEBUGPRINT("ok\n");
		}else{
			DEBUGPRINT("ng\n");
		}
		return ret;
	}

	//--------------------------------------
	//! @brief  １手すすめる
	//--------------------------------------
	void engine::playmove(int i, int j, COLOR color)
	{
		DEBUGPRINT("--- engine::playmove(i=%d,j=%d,color=%s)\n", i, j, colorstr(color) );
		assert( m_pImpl );

		//石を置く
		m_pImpl->SetStone( COLOR2STATE(color), i+1, j+1 );
		
		//記録
		m_pImpl->RecordMove( COLOR2STATE(color), i+1, j+1 );
	}
		
	//--------------------------------------
	//! @brief  １手戻す
	//--------------------------------------
	void engine::undo()
	{
		DEBUGPRINT("--- engine::undo()\n" );
		assert( m_pImpl );
		
		//戻す
		m_pImpl->UndoMove();
	}

	//--------------------------------------
	//! @brief  盤面取得
	//	COLOR engine::get_board(int i, int j)
	//	GTPの showboard コマンドに対応します。
	//	指定の座標の盤情報を返して下さい。
	//--------------------------------------
	COLOR engine::get_board(int i, int j)
	{
		DEBUGPRINT("--- engine::get_board(i=%d,j=%d)\n",i,j );
		assert( m_pImpl );
		
		Impl::STATE s = m_pImpl->GetState( i+1, j+1 );
		switch( s ){
		case Impl::STATE_BLACK:
			DEBUGPRINT("BLACK\n" );
			return BLACK;
		case Impl::STATE_WHITE:
			DEBUGPRINT("WHITE\n" );
			return WHITE;
		default:
			assert(!"engine::get_board() error state");
		case Impl::STATE_SPACE:
			DEBUGPRINT("EMPTY\n" );
			return EMPTY;
		}
	}


	//--------------------------------------
	//! @brief  
	//void engine::compute_final_status()
	//	GTPの final_score に対応しています。
	//	試合終了後に呼出します。
	//	最終的な盤面情報を決定して下さい。
	//--------------------------------------
	void engine::compute_final_status() 
	{
		DEBUGPRINT("--- engine::compute_final_status()\n" );
		assert( m_pImpl );
	}

	//--------------------------------------
	//! @brief  
	//COLOR engine::get_final_status(int i, int j)
	//	指定の座標の最終的な盤情報を返して下さい。
	//--------------------------------------
	COLOR engine::get_final_status(int i, int j)
	{
		DEBUGPRINT("--- engine::get_final_status(i=%d,j=%d)\n",i,j );
		assert( m_pImpl );

		Impl::STATE s = m_pImpl->GetState( i+1, j+1 );
		switch( s ){
		case Impl::STATE_BLACK:
			DEBUGPRINT("BLACK_TERRITORY\n" );
			return BLACK_TERRITORY;
		case Impl::STATE_WHITE:
			DEBUGPRINT("WHITE_TERRITORY\n" );
			return WHITE_TERRITORY;
		default:
			DEBUGPRINT("EMPTY\n" );
			return EMPTY;
		}
	}

};//namespace my_go {


